@@ -1,3 +1,4 @@
+
 #define G_LOG_DOMAIN "phoc-output"
 
 #include "phoc-config.h"
@@ -349,14 +350,24 @@ update_output_manager_config (PhocDesktop *desktop)
   wlr_output_manager_v1_set_configuration (desktop->output_manager_v1, config);
 }
 
+
 static void
 phoc_output_handle_destroy (struct wl_listener *listener, void *data)
 {
   PhocOutput *self = wl_container_of (listener, self, output_destroy);
+  PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
 
   if (self->fullscreen_view)
     phoc_view_set_fullscreen (self->fullscreen_view, false, NULL);
 
+  wl_list_remove (&priv->request_state.link);
+  wl_list_remove (&priv->damage.link);
+  wl_list_remove (&priv->frame.link);
+  wl_list_remove (&priv->needs_frame.link);
+  wl_list_remove (&self->commit.link);
+  wl_list_remove (&self->output_destroy.link);
+
+  /* This will trigger dispose */
   g_signal_emit (self, signals[OUTPUT_DESTROY], 0);
 }
 
@@ -387,8 +398,7 @@ phoc_output_handle_damage (struct wl_listener *listener, void *user_data)
   PhocOutput *self = PHOC_OUTPUT_SELF (priv);
   struct wlr_output_event_damage *event = user_data;
 
-  if (wlr_damage_ring_add (&self->damage_ring, event->damage))
-    wlr_output_schedule_frame (self->wlr_output);
+  phoc_output_damage_region (self, event->damage);
 }
 
 
@@ -501,21 +511,6 @@ scan_out_fullscreen_view (PhocOutput *self, PhocView *view, struct wlr_output_st
 }
 
 
-static void
-get_frame_damage (PhocOutput *self, pixman_region32_t *frame_damage)
-{
-  int width, height;
-  enum wl_output_transform transform;
-
-  wlr_output_transformed_resolution (self->wlr_output, &width, &height);
-
-  pixman_region32_init (frame_damage);
-
-  transform = wlr_output_transform_invert (self->wlr_output->transform);
-  wlr_region_transform (frame_damage, &self->damage_ring.current, transform, width, height);
-}
-
-
 static void
 build_debug_damage_tracking (PhocOutput *self)
 {
@@ -558,11 +553,7 @@ build_debug_damage_tracking (PhocOutput *self)
   };
 
   if (pixman_region32_not_empty (&highlight_damage)) {
-    gboolean intersects;
-
-    intersects = wlr_damage_ring_add (&self->damage_ring, &highlight_damage);
-    if (!intersects)
-      g_warning_once ("Damage not on output %p", self);
+    wlr_damage_ring_add (&self->damage_ring, &highlight_damage);
   }
 
   pixman_region32_fini (&highlight_damage);
@@ -580,6 +571,7 @@ phoc_output_draw (PhocOutput *self)
   struct wlr_buffer *buffer;
   struct wlr_render_pass *render_pass;
   struct wlr_output_state pending = { 0 };
+  PhocServerDebugFlags flags;
 
   if (!wlr_output->enabled)
     return;
@@ -595,12 +587,13 @@ phoc_output_draw (PhocOutput *self)
   if (G_UNLIKELY (priv->gamma_lut_changed))
     phoc_output_set_gamma_lut (self, &pending);
 
-  get_frame_damage (self, &frame_damage);
+  pixman_region32_init (&frame_damage);
+  pixman_region32_copy (&frame_damage, &self->damage_ring.current);
   wlr_output_state_set_damage (&pending, &frame_damage);
   pixman_region32_fini (&frame_damage);
 
   /* Check if we can delegate the fullscreen surface to the output */
-  if (phoc_output_has_fullscreen_view (self))
+  if (self->fullscreen_view)
     scanned_out = scan_out_fullscreen_view (self, self->fullscreen_view, &pending);
 
   if (scanned_out)
@@ -609,7 +602,7 @@ phoc_output_draw (PhocOutput *self)
   if (!wlr_output_configure_primary_swapchain (wlr_output, &pending, &wlr_output->swapchain))
     goto out;
 
-  buffer = wlr_swapchain_acquire (wlr_output->swapchain, NULL);
+  buffer = wlr_swapchain_acquire (wlr_output->swapchain);
   if (!buffer)
     goto out;
 
@@ -647,6 +640,10 @@ phoc_output_draw (PhocOutput *self)
 
  out:
   wlr_output_state_finish (&pending);
+
+  flags = phoc_server_get_debug_flags (phoc_server_get_default ());
+  if (G_UNLIKELY (flags & PHOC_SERVER_DEBUG_FLAG_DAMAGE_WHOLE))
+    phoc_output_damage_whole (self);
 }
 
 
@@ -676,8 +673,7 @@ phoc_output_handle_frame (struct wl_listener *listener, void *data)
   /* Ensure the cutouts are drawn */
   if (G_UNLIKELY (priv->cutouts_texture)) {
     struct wlr_box box = { 0, 0, priv->cutouts_texture->width, priv->cutouts_texture->height };
-    if (wlr_damage_ring_add_box (&self->damage_ring, &box))
-      wlr_output_schedule_frame (self->wlr_output);
+    phoc_output_damage_box (self, &box);
   }
 
   build_debug_damage_tracking (self);
@@ -754,9 +750,6 @@ phoc_output_handle_commit (struct wl_listener *listener, void *data)
 
   if (event->state->committed & (WLR_OUTPUT_STATE_MODE |
                                  WLR_OUTPUT_STATE_TRANSFORM)) {
-    int width, height;
-    wlr_output_transformed_resolution (self->wlr_output, &width, &height);
-    wlr_damage_ring_set_bounds (&self->damage_ring, width, height);
     wlr_output_schedule_frame (self->wlr_output);
   }
 
@@ -915,10 +908,12 @@ phoc_output_fill_state (PhocOutput              *self,
     if (output_config->phys_height)
       self->wlr_output->phys_height = output_config->phys_height;
 
-    if (output_config->mode.width)
+    if (output_config->mode.width &&
+        output_config->mode.height) {
       phoc_output_state_set_mode (self, pending, output_config);
-    else if (preferred_mode != NULL)
+    } else if (preferred_mode != NULL) {
       wlr_output_state_set_mode (pending, preferred_mode);
+    }
 
     if (output_config->scale)
       scale = output_config->scale;
@@ -929,6 +924,13 @@ phoc_output_fill_state (PhocOutput              *self,
 
     wlr_output_state_set_transform (pending, transform);
     priv->scale_filter = output_config->scale_filter;
+
+    if (output_config->adaptive_sync != PHOC_OUTPUT_ADAPTIVE_SYNC_NONE &&
+        self->wlr_output->adaptive_sync_supported) {
+      bool enabled = output_config->adaptive_sync == PHOC_OUTPUT_ADAPTIVE_SYNC_ENABLED;
+
+      wlr_output_state_set_adaptive_sync_enabled (pending, enabled);
+    }
   } else if (enable) {
     enum wl_output_transform transform = WL_OUTPUT_TRANSFORM_NORMAL;
     gboolean has_mode = FALSE;
@@ -961,13 +963,26 @@ phoc_output_fill_state (PhocOutput              *self,
     wlr_output_state_set_scale (pending, phoc_output_compute_scale (self, pending));
     wlr_output_state_set_transform (pending, transform);
   }
+}
 
-  if (output_config && output_config->x > 0 && output_config->y > 0) {
-    wlr_output_layout_add (self->desktop->layout, self->wlr_output, output_config->x,
+
+static void
+phoc_output_set_layout_pos (PhocOutput *self, PhocOutputConfig *output_config)
+{
+  struct wlr_box output_box;
+
+  if (output_config && output_config->x >= 0 && output_config->y >= 0) {
+    wlr_output_layout_add (self->desktop->layout,
+                           self->wlr_output,
+                           output_config->x,
                            output_config->y);
   } else {
     wlr_output_layout_add_auto (self->desktop->layout, self->wlr_output);
   }
+
+  wlr_output_layout_get_box (self->desktop->layout, self->wlr_output, &output_box);
+  self->lx = output_box.x;
+  self->ly = output_box.y;
 }
 
 
@@ -977,15 +992,16 @@ phoc_output_initable_init (GInitable    *initable,
                            GError      **error)
 {
   PhocServer *server = phoc_server_get_default ();
+  PhocDesktop *desktop = phoc_server_get_desktop (server);
   PhocInput *input = phoc_server_get_input (server);
   PhocRenderer *renderer = phoc_server_get_renderer (server);
+  PhocConfig *config = phoc_server_get_config (phoc_server_get_default ());
   PhocOutput *self = PHOC_OUTPUT (initable);
   PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
-  struct wlr_box output_box;
+  PhocOutputConfig *output_config;
+  struct wlr_output_state pending;
   int width, height;
 
-  PhocConfig *config = phoc_server_get_config (phoc_server_get_default ());
-
   self->wlr_output->data = self;
   wl_list_insert (&self->desktop->outputs, &self->link);
 
@@ -1018,14 +1034,22 @@ phoc_output_initable_init (GInitable    *initable,
   priv->request_state.notify = handle_request_state;
   wl_signal_add (&self->wlr_output->events.request_state, &priv->request_state);
 
-  PhocOutputConfig *output_config = phoc_config_get_output (config, self);
-  struct wlr_output_state pending;
-  phoc_output_fill_state (self, output_config, &pending);
+  output_config = phoc_config_get_output (config, self);
+  /* Restore old output state if any */
+  if (output_config) {
+    phoc_output_config_dump (output_config, "Config file ");
+  } else {
+    const char *identifier = phoc_output_get_identifier (self);
 
+    output_config = phoc_desktop_get_saved_outputs_state (desktop, identifier);
+    if (output_config) {
+      g_message ("Loading saved output state for '%s'", identifier);
+      phoc_output_config_dump (output_config, "Saved output state ");
+    }
+  }
+  phoc_output_fill_state (self, output_config, &pending);
+  phoc_output_set_layout_pos (self, output_config);
   wlr_output_commit_state (self->wlr_output, &pending);
-  wlr_output_layout_get_box (self->desktop->layout, self->wlr_output, &output_box);
-  self->lx = output_box.x;
-  self->ly = output_box.y;
 
   for (GSList *elem = phoc_input_get_seats (input); elem; elem = elem->next) {
     PhocSeat *seat = PHOC_SEAT (elem->data);
@@ -1042,7 +1066,6 @@ phoc_output_initable_init (GInitable    *initable,
   update_output_manager_config (self->desktop);
 
   wlr_output_transformed_resolution (self->wlr_output, &width, &height);
-  wlr_damage_ring_set_bounds (&self->damage_ring, width, height);
 
   if (phoc_server_check_debug_flags (server, PHOC_SERVER_DEBUG_FLAG_CUTOUTS)) {
     priv->cutouts = phoc_cutouts_overlay_new (phoc_server_get_compatibles (server));
@@ -1084,13 +1107,6 @@ phoc_output_finalize (GObject *object)
 
   update_output_manager_config (self->desktop);
 
-  wl_list_remove (&self->commit.link);
-  wl_list_remove (&self->output_destroy.link);
-
-  wl_list_remove (&priv->request_state.link);
-  wl_list_remove (&priv->damage.link);
-  wl_list_remove (&priv->frame.link);
-  wl_list_remove (&priv->needs_frame.link);
   wlr_damage_ring_finish (&self->damage_ring);
 
   /* Remove all frame callbacks, this will also free associated user data */
@@ -1522,6 +1538,25 @@ phoc_output_drag_icons_for_each_surface (PhocOutput          *self,
   }
 }
 
+
+struct for_each_surface_data {
+  PhocOutput          *output;
+  gboolean             visible_only;
+  PhocSurfaceIterator  iterator;
+  gpointer             user_data;
+};
+
+static gboolean
+for_each_surface_iter (PhocDesktop *desktop, PhocView *view, gpointer user_data)
+{
+  struct for_each_surface_data *data = user_data;
+
+  if (!data->visible_only || phoc_desktop_view_check_visibility (desktop, view))
+    phoc_output_view_for_each_surface (data->output, view, data->iterator, data->user_data);
+
+  return TRUE;
+}
+
 /**
  * phoc_output_for_each_surface:
  * @self: the output
@@ -1553,12 +1588,14 @@ phoc_output_for_each_surface (PhocOutput          *self,
     }
 #endif
   } else {
-    for (GList *l = phoc_desktop_get_views (desktop)->tail; l; l = l->prev) {
-      PhocView *view = PHOC_VIEW (l->data);
-
-      if (!visible_only || phoc_desktop_view_check_visibility (desktop, view))
-        phoc_output_view_for_each_surface (self, view, iterator, user_data);
-    }
+    phoc_desktop_for_each_view (desktop,
+                                for_each_surface_iter,
+                                (gpointer)&(struct for_each_surface_data){
+                                  .output = self,
+                                  .iterator = iterator,
+                                  .user_data = user_data,
+                                  .visible_only = visible_only,
+                                });
   }
 
   phoc_output_drag_icons_for_each_surface (self, input, iterator, user_data);
@@ -1633,14 +1670,11 @@ damage_surface_iterator (PhocOutput *self, struct wlr_surface *wlr_surface, stru
   }
 
   pixman_region32_translate (&damage, box.x, box.y);
-  if (wlr_damage_ring_add (&self->damage_ring, &damage))
-    wlr_output_schedule_frame (self->wlr_output);
+  phoc_output_damage_region (self, &damage);
   pixman_region32_fini (&damage);
 
-  if (*whole) {
-    if (wlr_damage_ring_add_box (&self->damage_ring, &box))
-      wlr_output_schedule_frame (self->wlr_output);
-  }
+  if (*whole)
+    phoc_output_damage_box (self, &box);
 
   if (!wl_list_empty (&wlr_surface->current.frame_callback_list))
     wlr_output_schedule_frame (self->wlr_output);
@@ -1648,14 +1682,13 @@ damage_surface_iterator (PhocOutput *self, struct wlr_surface *wlr_surface, stru
 
 
 static void
-damage_whole_view (PhocOutput *self, PhocView  *view)
+damage_view_blings (PhocOutput *self, PhocView  *view)
 {
   GSList *blings;
   struct wlr_box box;
 
-  if (!phoc_view_is_mapped (view)) {
+  if (!phoc_view_is_mapped (view))
     return;
-  }
 
   blings = phoc_view_get_blings (view);
   if (G_LIKELY (!blings))
@@ -1669,8 +1702,7 @@ damage_whole_view (PhocOutput *self, PhocView  *view)
     box.y -= self->ly;
     phoc_utils_scale_box (&box, self->wlr_output->scale);
 
-    if (wlr_damage_ring_add_box (&self->damage_ring, &box))
-      wlr_output_schedule_frame (self->wlr_output);
+    phoc_output_damage_box (self, &box);
   }
 }
 
@@ -1692,7 +1724,7 @@ phoc_output_damage_from_view (PhocOutput *self, PhocView *view, bool whole)
   }
 
   if (whole)
-    damage_whole_view (self, view);
+    damage_view_blings (self, view);
 
   phoc_output_view_for_each_surface (self, view, damage_surface_iterator, &whole);
 }
@@ -1759,18 +1791,122 @@ phoc_output_damage_from_surface (PhocOutput         *self,
                                         damage_surface_iterator, &whole);
 }
 
+/**
+ * phoc_output_damage_region:
+ * @self: The output
+ * @region: The damage in output local coordinates
+ *
+ * If damage overlaps with output add it and schedule a frame.
+ *
+ * Returns: `TRUE` if the damage overlapped with the output
+ */
+gboolean
+phoc_output_damage_region (PhocOutput *self, const pixman_region32_t *region)
+{
+  pixman_region32_t clipped;
+  int width, height;
 
-static void
+  wlr_output_transformed_resolution (self->wlr_output, &width, &height);
+
+  pixman_region32_init (&clipped);
+  pixman_region32_intersect_rect (&clipped, region, 0, 0, width, height);
+
+  if (!pixman_region32_not_empty (&clipped)) {
+    pixman_region32_fini (&clipped);
+    return FALSE;
+  }
+
+  /* Transform to damage ring buffer local coordinates */
+  phoc_output_transform_damage (self, &clipped);
+  wlr_damage_ring_add (&self->damage_ring, &clipped);
+
+  pixman_region32_fini (&clipped);
+  wlr_output_schedule_frame (self->wlr_output);
+  return TRUE;
+}
+
+/**
+ * phoc_output_damage_box:
+ * @self: The output
+ * @box: The damage box in output local coordinates
+ *
+ * If damage overlaps with output add it and schedule a frame.
+ *
+ * Returns: `TRUE` if the damage overlapped with the output
+ */
+gboolean
+phoc_output_damage_box (PhocOutput *self, const struct wlr_box *box)
+{
+  struct wlr_box clipped;
+  int width, height;
+
+  wlr_output_transformed_resolution (self->wlr_output, &width, &height);
+
+  clipped = (struct wlr_box) { .x = 0, .y = 0, .width = width, .height = height };
+  if (!wlr_box_intersection (&clipped, &clipped, box))
+    return FALSE;
+
+  /* Transform to damage ring buffer local coordinates */
+  phoc_output_transform_box (self, &clipped);
+  wlr_damage_ring_add_box (&self->damage_ring, &clipped);
+
+  wlr_output_schedule_frame (self->wlr_output);
+  return TRUE;
+}
+
+
+static PhocOutputConfig *
+phoc_output_config_head_to_output_config (PhocOutput                              *self,
+                                          struct wlr_output_configuration_head_v1 *head)
+{
+  PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
+  PhocOutputConfig *oc = phoc_output_config_new (phoc_output_get_identifier (self));
+
+  oc->enable = TRUE;
+  if (head->state.mode && !head->state.mode->preferred) {
+    oc->mode.width = head->state.mode->width;
+    oc->mode.height = head->state.mode->height;
+    oc->mode.refresh_rate = head->state.mode->refresh / 1000.0;
+  } else {
+    oc->mode.width = head->state.custom_mode.width;
+    oc->mode.height = head->state.custom_mode.height;
+    oc->mode.refresh_rate = head->state.custom_mode.refresh / 1000.0;
+  }
+
+  oc->transform = head->state.transform;
+  oc->scale = adjust_frac_scale (head->state.scale);
+  oc->scale_filter = priv->scale_filter;
+
+  if (self->wlr_output->adaptive_sync_supported) {
+    if (head->state.adaptive_sync_enabled)
+      oc->adaptive_sync = PHOC_OUTPUT_ADAPTIVE_SYNC_ENABLED;
+    else
+      oc->adaptive_sync = PHOC_OUTPUT_ADAPTIVE_SYNC_DISABLED;
+  }
+
+  oc->x = head->state.x;
+  oc->y = head->state.y;
+
+  phoc_output_config_dump (oc, "Head state ");
+
+  return oc;
+}
+
+
+static gboolean
 output_manager_apply_config (PhocDesktop                        *desktop,
-                             struct wlr_output_configuration_v1 *config,
-                             gboolean                            test_only)
-
+                             struct wlr_output_configuration_v1 *wlr_config_v1,
+                             gboolean                            test_only,
+                             GPtrArray                         **out_configs)
 {
   struct wlr_output_configuration_head_v1 *config_head;
   gboolean ok = TRUE;
+  g_autoptr (GPtrArray) output_configs = NULL;
+
+  output_configs = g_ptr_array_new_full (5, (GDestroyNotify) phoc_output_config_destroy);
 
   /* First disable outputs we need to disable */
-  wl_list_for_each (config_head, &config->heads, link) {
+  wl_list_for_each (config_head, &wlr_config_v1->heads, link) {
     struct wlr_output *wlr_output = config_head->state.output;
     struct wlr_output_state pending;
 
@@ -1792,57 +1928,48 @@ output_manager_apply_config (PhocDesktop                        *desktop,
   }
 
   /* Then enable outputs that need to */
-  wl_list_for_each (config_head, &config->heads, link) {
+  wl_list_for_each (config_head, &wlr_config_v1->heads, link) {
     struct wlr_output *wlr_output = config_head->state.output;
     PhocOutput *output = PHOC_OUTPUT (wlr_output->data);
     struct wlr_output_state pending;
-    struct wlr_box output_box;
+    g_autoptr (PhocOutputConfig) oc = NULL;
 
     if (!config_head->state.enabled)
       continue;
 
-    wlr_output_state_init (&pending);
-    wlr_output_state_set_enabled (&pending, true);
-    if (config_head->state.mode != NULL) {
-      wlr_output_state_set_mode (&pending, config_head->state.mode);
-    } else {
-      wlr_output_state_set_custom_mode (&pending,
-                                        config_head->state.custom_mode.width,
-                                        config_head->state.custom_mode.height,
-                                        config_head->state.custom_mode.refresh);
-    }
-    wlr_output_state_set_transform (&pending, config_head->state.transform);
-    wlr_output_state_set_scale (&pending, adjust_frac_scale (config_head->state.scale));
+    oc = phoc_output_config_head_to_output_config (output, config_head);
+    phoc_output_fill_state (output, oc, &pending);
 
     if (test_only) {
       ok &= wlr_output_test_state (wlr_output, &pending);
     } else {
-      wlr_output_layout_add (desktop->layout,
-                             wlr_output,
-                             config_head->state.x,
-                             config_head->state.y);
       ok &= wlr_output_commit_state (wlr_output, &pending);
 
+      phoc_output_set_layout_pos (output, oc);
+
       if (output->fullscreen_view)
         phoc_view_set_fullscreen (output->fullscreen_view, true, output);
-
-      wlr_output_layout_get_box (output->desktop->layout, output->wlr_output, &output_box);
-      output->lx = output_box.x;
-      output->ly = output_box.y;
     }
 
     wlr_output_state_finish (&pending);
+
+    g_ptr_array_add (output_configs, g_steal_pointer (&oc));
   }
 
   if (ok)
-    wlr_output_configuration_v1_send_succeeded (config);
+    wlr_output_configuration_v1_send_succeeded (wlr_config_v1);
   else
-    wlr_output_configuration_v1_send_failed (config);
+    wlr_output_configuration_v1_send_failed (wlr_config_v1);
 
-  wlr_output_configuration_v1_destroy (config);
+  wlr_output_configuration_v1_destroy (wlr_config_v1);
 
   if (!test_only)
     update_output_manager_config (desktop);
+
+  if (out_configs)
+    *out_configs = g_steal_pointer (&output_configs);
+
+  return ok;
 }
 
 
@@ -1851,8 +1978,13 @@ phoc_handle_output_manager_apply (struct wl_listener *listener, void *data)
 {
   PhocDesktop *desktop = wl_container_of (listener, desktop, output_manager_apply);
   struct wlr_output_configuration_v1 *config = data;
+  gboolean success;
+  g_autoptr (GPtrArray) output_configs = NULL;
 
-  output_manager_apply_config (desktop, config, FALSE);
+  success = output_manager_apply_config (desktop, config, FALSE, &output_configs);
+
+  if (success)
+    phoc_desktop_save_outputs_state (desktop, g_steal_pointer (&output_configs));
 }
 
 
@@ -1862,7 +1994,7 @@ phoc_handle_output_manager_test (struct wl_listener *listener, void *data)
   PhocDesktop *desktop = wl_container_of (listener, desktop, output_manager_apply);
   struct wlr_output_configuration_v1 *config = data;
 
-  output_manager_apply_config (desktop, config, TRUE);
+  output_manager_apply_config (desktop, config, TRUE, NULL);
 }
 
 
@@ -1963,21 +2095,6 @@ phoc_output_is_match (PhocOutput *self,
   return match;
 }
 
-/**
- * phoc_output_has_fullscreen_view:
- * @self: The #PhocOutput
- *
- * Returns: %TRUE if the output has a fullscreen view attached,
- *          %FALSE otherwise.
- */
-gboolean
-phoc_output_has_fullscreen_view (PhocOutput *self)
-{
-  g_assert (PHOC_IS_OUTPUT (self));
-
-  return phoc_view_is_mapped (self->fullscreen_view);
-}
-
 
 guint
 phoc_output_add_frame_callback  (PhocOutput        *self,
@@ -2284,6 +2401,19 @@ phoc_output_get_name (PhocOutput *self)
 }
 
 
+const char *
+phoc_output_get_identifier (PhocOutput *self)
+{
+  g_assert (PHOC_IS_OUTPUT (self));
+  g_assert (self->wlr_output);
+
+  if (self->wlr_output->description)
+    return self->wlr_output->description;
+
+  return self->wlr_output->name;
+}
+
+
 void
 phoc_output_handle_gamma_control_set_gamma (struct wl_listener *listener, void *data)
 {
