@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2021 Purism SPC
- *               2023-2024 The Phosh Developers
+ *               2023-2025 The Phosh Developers
  *
  * SPDX-License-Identifier: GPL-3.0-or-later or MIT
  */
@@ -298,7 +298,7 @@ handle_tablet_tool_position (PhocCursor             *cursor,
     return;
   }
 
-  if (!wlr_surface_accepts_tablet_v2 (tablet->tablet_v2, surface)) {
+  if (!wlr_surface_accepts_tablet_v2 (surface, tablet->tablet_v2)) {
     wlr_tablet_v2_tablet_tool_notify_proximity_out (phoc_tool->tablet_v2_tool);
     /* XXX: TODO: Fallback pointer semantics */
     return;
@@ -1377,6 +1377,9 @@ phoc_seat_allow_input (PhocSeat *seat, struct wl_resource *resource)
   g_assert (PHOC_IS_SEAT (seat));
   priv = phoc_seat_get_instance_private (seat);
 
+  if (!phoc_server_get_allow_input (phoc_server_get_default ()))
+    return false;
+
   return !priv->exclusive_client || wl_resource_get_client (resource) == priv->exclusive_client;
 }
 
@@ -1466,7 +1469,7 @@ phoc_seat_set_focus_view (PhocSeat *seat, PhocView *view)
   if (view && PHOC_IS_XWAYLAND_SURFACE (view)) {
     struct wlr_xwayland_surface *xsurface =
       phoc_xwayland_surface_get_wlr_surface (PHOC_XWAYLAND_SURFACE (view));
-    if (!wlr_xwayland_or_surface_wants_focus (xsurface))
+    if (!wlr_xwayland_surface_override_redirect_wants_focus (xsurface))
       return;
   }
 #endif
@@ -1875,6 +1878,7 @@ phoc_seat_finalize (GObject *object)
   wlr_seat_destroy (self->seat);
   g_clear_pointer (&priv->name, g_free);
 
+  g_clear_pointer (&self->keyboard_shortcuts_inhibitors, g_slist_free);
   g_clear_pointer (&self->keyboards, g_slist_free);
   g_clear_pointer (&self->pointers, g_slist_free);
   g_clear_pointer (&self->switches, g_slist_free);
@@ -2078,3 +2082,42 @@ phoc_seat_get_last_event_ts (PhocSeat *self)
 
   return priv->last_event_ts;
 }
+
+
+gboolean
+phoc_seat_shortcuts_inhibited (const PhocSeat *self)
+{
+  struct wlr_surface *focused_surface = self->seat->keyboard_state.focused_surface;
+  PhocKeyboardShortcutsInhibit* inhibit;
+
+  for (GSList *l = self->keyboard_shortcuts_inhibitors; l; l = l->next) {
+    inhibit = PHOC_KEYBOARD_SHORTCUTS_INHIBIT (l->data);
+    if (phoc_keyboard_shortcuts_inhibit_get_surface (inhibit) == focused_surface)
+      return TRUE;
+  }
+  return FALSE;
+}
+
+
+static void
+on_shortcuts_inhibit_destroy (PhocSeat *self, PhocKeyboardShortcutsInhibit* inhibit)
+{
+  self->keyboard_shortcuts_inhibitors = g_slist_remove (self->keyboard_shortcuts_inhibitors,
+                                                        inhibit);
+  g_object_unref (inhibit);
+}
+
+
+void
+phoc_seat_add_shortcuts_inhibit (PhocSeat                                   *self,
+                                 struct wlr_keyboard_shortcuts_inhibitor_v1 *inhibitor)
+{
+  PhocKeyboardShortcutsInhibit *inhibit;
+  inhibit = phoc_keyboard_shortcuts_inhibit_new (inhibitor);
+
+  self->keyboard_shortcuts_inhibitors = g_slist_prepend (self->keyboard_shortcuts_inhibitors,
+                                                         inhibit);
+  g_signal_connect_swapped (inhibit, "destroy",
+                            G_CALLBACK (on_shortcuts_inhibit_destroy),
+                            self);
+}
